# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Read the detailed environmental conditions of a thermal mapping analysis from
the env_conds output by a thermal mapping component.
_
Environemntal conditions include raw inputs to the thermal comfort model, such as
air temperature, MRT, longwave MRT, and shortwave MRT delta.

-
    Args:
        _env_conds: Path to a folder containing the detailed environmental conditions
            output by a thermal mapping component.
        _metric_: Text or an integer for the specific metric to be loaded from the
            environmental conditions. (Default: MRT). Choose from the following.
                * 0 - MRT
                * 1 - Air Temperature
                * 2 - Longwave MRT
                * 3 - Shortwave MRT Delta
                * 4 - Relative Humidity
        _load: Set to True to load the data into Grasshopper.

    Returns:
        comf_mtx: A Matrix object that can be connected to the "HB Visualize Thermal
            Map" component in order to spatially visualize results. This Matrix
            object can also be connected to the "LB Deconstruct Matrix"
            component to obtain detailed point-by-point and hour-by-hour
            values.
            _
            When deconstructed, each sub-list of the matrix (aka. branch of the
            Data Tree) represents one of the sensor grids used for analysis.
            The length of each sub-list matches the number of points in the
            grid. Each value in the sub-list is an hourly data collection
            containing hour-by-hour results for each point.
"""

ghenv.Component.Name = 'HB Read Environment Matrix'
ghenv.Component.NickName = 'EnvMtx'
ghenv.Component.Message = '1.6.0'
ghenv.Component.Category = 'HB-Energy'
ghenv.Component.SubCategory = '7 :: Thermal Map'
ghenv.Component.AdditionalHelpFromDocStrings = '0'

import subprocess
import os
import shutil
import json

try:
    from ladybug.datatype.temperature import AirTemperature, \
        MeanRadiantTemperature, RadiantTemperature
    from ladybug.datatype.temperaturedelta import RadiantTemperatureDelta
    from ladybug.datatype.fraction import RelativeHumidity
    from ladybug.header import Header
    from ladybug.datacollection import HourlyContinuousCollection, \
        HourlyDiscontinuousCollection
    from ladybug.futil import csv_to_num_matrix
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:
    from honeybee.config import folders
except ImportError as e:
    raise ImportError('\nFailed to import honeybee:\n\t{}'.format(e))

try:
    from ladybug_rhino.grasshopper import all_required_inputs, objectify_output
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))

ENV_CONDS_MAP = {
    '0': 'mrt',
    'mrt': 'mrt',
    'mean radiant temperature': 'mrt',
    '1': 'air_temperature',
    'air temperature': 'air_temperature',
    '2': 'longwave_mrt',
    'longwave mrt': 'longwave_mrt',
    '3': 'shortwave_mrt',
    'shortwave mrt': 'shortwave_mrt',
    'shortwave mrt delta': 'shortwave_mrt',
    '4': 'rel_humidity',
    'relative humidity':  'rel_humidity'
}


def load_matrix(comf_result):
    """Load a matrix of data into an object that can be output in Grasshopper.

    Args:
        comf_result: Path to a folder with CSV data to be loaded into Grasshopper.
    """
    # parse the result_info.json into a data collection header
    with open(os.path.join(comf_result, 'results_info.json')) as json_file:
        data_header = Header.from_dict(json.load(json_file))
    a_per = data_header.analysis_period
    continuous = True if a_per.st_hour == 0 and a_per.end_hour == 23 else False
    if not continuous:
        dates = a_per.datetimes

    # parse the grids_info.json with the correct order of the grid files
    with open(os.path.join(comf_result, 'grids_info.json')) as json_file:
        grid_list = json.load(json_file)

    # loop through the grid CSV files, parse their results, and build data collections
    comf_matrix = []
    for grid in grid_list:
        grid_name = grid['full_id'] if 'full_id' in grid else 'id'
        metadata = {'grid': grid_name}
        grid_file = os.path.join(comf_result, '{}.csv'.format(grid_name))
        data_matrix = csv_to_num_matrix(grid_file)
        grid_data = []
        for i, row in enumerate(data_matrix):
            header = data_header.duplicate()
            header.metadata = metadata.copy()
            header.metadata['sensor_index'] = i
            data = HourlyContinuousCollection(header, row) if continuous else \
                HourlyDiscontinuousCollection(header, row, dates)
            grid_data.append(data)
        comf_matrix.append(grid_data)

    # wrap the maptrix into an object so that it does not slow the Grasshopper UI
    comf_mtx = objectify_output(
        '{} Matrix'.format(data_header.data_type.name), comf_matrix)
    return comf_mtx


def create_result_header(env_conds, sub_path):
    """Create a DataCollection Header for a given metric."""
    with open(os.path.join(env_conds, 'results_info.json')) as json_file:
        base_head = Header.from_dict(json.load(json_file))
    if sub_path == 'mrt':
        return Header(MeanRadiantTemperature(), 'C', base_head.analysis_period)
    elif sub_path == 'air_temperature':
        return Header(AirTemperature(), 'C', base_head.analysis_period)
    elif sub_path == 'longwave_mrt':
        return Header(RadiantTemperature(), 'C', base_head.analysis_period)
    elif sub_path == 'shortwave_mrt':
        return Header(RadiantTemperatureDelta(), 'dC', base_head.analysis_period)
    elif sub_path == 'rel_humidity':
        return Header(RelativeHumidity(), '%', base_head.analysis_period)


def sum_matrices(mtxs_1, mtxs_2, dest_dir):
    """Sum together matrices of two folders."""
    if not os.path.isdir(dest_dir):
        os.makedirs(dest_dir)
    for mtx_file in os.listdir(mtxs_1):
        if mtx_file.endswith('.csv'):
            mtx_file1 = os.path.join(mtxs_1, mtx_file)
            mtx_file2 = os.path.join(mtxs_2, mtx_file)
            matrix_1 = csv_to_num_matrix(mtx_file1)
            matrix_2 = csv_to_num_matrix(mtx_file2)
            data = [[d1 + d2 for d1, d2 in zip(r1, r2)]
                    for r1, r2 in zip(matrix_1, matrix_2)]
            csv_path = os.path.join(dest_dir, mtx_file)
            with open(csv_path, 'w') as csv_file:
                for dat in data:
                    str_data = (str(v) for v in dat)
                    csv_file.write(','.join(str_data) + '\n')
        elif mtx_file == 'grids_info.json':
            shutil.copyfile(
                os.path.join(mtxs_1, mtx_file),
                os.path.join(dest_dir, mtx_file)
            )


if all_required_inputs(ghenv.Component) and _load:
    # get the folders and that correspond with the requested metric
    _metric_ = _metric_ if _metric_ is not None else 'mrt'
    try:
        sub_path = ENV_CONDS_MAP[_metric_.lower()]
    except KeyError:
        raise ValueError(
            'Input metric "{}" is not recognized. Choose from: {}'.format(
                _metric_, '\n'.join(ENV_CONDS_MAP.keys()))
        )
    source_folder = os.path.join(_env_conds, sub_path)
    dest_folder = os.path.join(_env_conds, 'final', sub_path)

    # if the results have already been processed, then load them up
    if os.path.isdir(dest_folder):
        comf_mtx = load_matrix(dest_folder)
    else:  # otherwise, process them into a load-able format
        # make sure the requested metric is valid for the study
        if sub_path == 'mrt':
            source_folders = [os.path.join(_env_conds, 'longwave_mrt'),
                              os.path.join(_env_conds, 'shortwave_mrt')]
            dest_folders = [os.path.join(_env_conds, 'final', 'longwave_mrt'),
                            os.path.join(_env_conds, 'final', 'shortwave_mrt')]
        else:
            assert os.path.isdir(source_folder), \
                'Metric "{}" does not exist for this comfort study.'.format(sub_path)
            source_folders, dest_folders = [source_folder], [dest_folder]
        # restructure the results to align with the sensor grids
        dist_info = os.path.join(_env_conds, '_redist_info.json')
        for src_f, dst_f in zip(source_folders, dest_folders):
            if not os.path.isdir(dst_f):
                os.makedirs(dst_f)
                cmds = [folders.python_exe_path, '-m', 'honeybee_radiance', 'grid',
                        'merge-folder', src_f, dst_f, 'csv',
                        '--dist-info', dist_info]
                shell = True if os.name == 'nt' else False
                process = subprocess.Popen(cmds, stdout=subprocess.PIPE, shell=shell)
                stdout = process.communicate()
                grid_info_src = os.path.join(_env_conds, 'grids_info.json')
                grid_info_dst = os.path.join(dst_f, 'grids_info.json')
                shutil.copyfile(grid_info_src, grid_info_dst)
            data_header = create_result_header(_env_conds, os.path.split(dst_f)[-1])
            result_info_path = os.path.join(dst_f, 'results_info.json')
            with open(result_info_path, 'w') as fp:
                json.dump(data_header.to_dict(), fp, indent=4)
        # if MRT was requested, sum together the longwave and shortwave
        if sub_path == 'mrt':
            sum_matrices(dest_folders[0], dest_folders[1], dest_folder)
            data_header = create_result_header(_env_conds, sub_path)
            result_info_path = os.path.join(dest_folder, 'results_info.json')
            with open(result_info_path, 'w') as fp:
                json.dump(data_header.to_dict(), fp, indent=4)
        # load the resulting matrix into Grasshopper
        comf_mtx = load_matrix(dest_folder)



# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Deconstruct a Ventilation object into its constituient properties.
_
Note the the 4 ventilation types (_flow_per_person_, _flow_per_area_, _flow_per_zone_,
_ach_) are ultimately summed together to yeild the ventilation design flow rate used
in the simulation.
-

    Args:
        _vent: An Ventilation object to be deconstructed.
    
    Returns:
        name: Text string for the ventilation display name.
        flow_per_person: A numerical value for the intensity of ventilation
            in m3/s per person. Note that setting this value here does not mean
            that ventilation is varied based on real-time occupancy but rather
            that the design level of ventilation is determined using this value
            and the People object of the zone. To vary ventilation in real time,
            the ventilation schedule should be used. Most ventilation standards
            support that a value of 0.01 m3/s (10 L/s or ~20 cfm) per person is
            sufficient to remove odors.
        flow_per_area: A numerical value for the intensity of ventilation in m3/s
            per square meter of floor area.
        flow_per_zone: A numerical value for the design level of ventilation
            in m3/s for the entire zone.
        ach: A numberical value for the design level of ventilation in air changes
            per hour (ACH) for the entire zone. This is particularly helpful
            for hospitals, where ventilation standards are often given in ACH.
        schedule: An optional fractional schedule for the ventilation over the
            course of the year. The fractional values will get multiplied by
            the total design flow rate (determined from the fields above and the
            calculation_method) to yield a complete ventilation profile. If None,
            the design level of ventilation is used throughout all timesteps of
            the simulation.
"""

ghenv.Component.Name = "HB Deconstruct Ventilation"
ghenv.Component.NickName = 'DecnstrVentilation'
ghenv.Component.Message = '1.6.0'
ghenv.Component.Category = 'HB-Energy'
ghenv.Component.SubCategory = '3 :: Loads'
ghenv.Component.AdditionalHelpFromDocStrings = "0"

try:
    from honeybee_energy.load.ventilation import Ventilation
except ImportError as e:
    raise ImportError('\nFailed to import honeybee_energy:\n\t{}'.format(e))

try:
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component) and _vent is not None:
    # check the input
    assert isinstance(_vent, Ventilation), \
        'Expected Ventilation object. Got {}.'.format(type(_vent))

    # get the properties of the object
    name = _vent.display_name
    flow_per_person = _vent.flow_per_person
    flow_per_area = _vent.flow_per_area
    flow_per_zone = _vent.flow_per_zone
    ach = _vent.air_changes_per_hour
    schedule = _vent.schedule



# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Parse all of the common Room-level comfort-related results from an SQL result
file that has been generated from an energy simulation.

-
    Args:
        _sql: The file path of the SQL result file that has been generated from
            an energy simulation.
    
    Returns:
        oper_temp: DataCollections for the mean operative temperature of each room (C).
        air_temp: DataCollections for the mean air temperature of each room (C).
        rad_temp: DataCollections for the mean radiant temperature of each room (C).
        rel_humidity: DataCollections for the relative humidity of each room (%).
"""

ghenv.Component.Name = 'HB Read Room Comfort Result'
ghenv.Component.NickName = 'RoomComfortResult'
ghenv.Component.Message = '1.6.0'
ghenv.Component.Category = 'HB-Energy'
ghenv.Component.SubCategory = '6 :: Result'
ghenv.Component.AdditionalHelpFromDocStrings = '1'

import os
import subprocess
import json

try:
    from ladybug.datacollection import HourlyContinuousCollection, \
        MonthlyCollection, DailyCollection
    from ladybug.sql import SQLiteResult
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:
    from honeybee.config import folders
except ImportError as e:
    raise ImportError('\nFailed to import honeybee:\n\t{}'.format(e))

try:
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


def serialize_data(data_dicts):
    """Reserialize a list of collection dictionaries."""
    if len(data_dicts) == 0:
        return []
    elif data_dicts[0]['type'] == 'HourlyContinuous':
        return [HourlyContinuousCollection.from_dict(data) for data in data_dicts]
    elif data_dicts[0]['type'] == 'Monthly':
        return [MonthlyCollection.from_dict(data) for data in data_dicts]
    elif data_dicts[0]['type'] == 'Daily':
        return [DailyCollection.from_dict(data) for data in data_dicts]


# List of all the output strings that will be requested
oper_temp_output = 'Zone Operative Temperature'
air_temp_output = 'Zone Mean Air Temperature'
rad_temp_output = 'Zone Mean Radiant Temperature'
rel_humidity_output = 'Zone Air Relative Humidity'
all_output = [oper_temp_output, air_temp_output, rad_temp_output, rel_humidity_output]


if all_required_inputs(ghenv.Component):
    # check the size of the SQL file to see if we should use the CLI
    assert os.path.isfile(_sql), 'No sql file found at: {}.'.format(_sql)
    if os.name == 'nt' and os.path.getsize(_sql) < 1e8:
        # small file on windows; use IronPython like usual
        sql_obj = SQLiteResult(_sql)  # create the SQL result parsing object
        # get all of the results
        oper_temp = sql_obj.data_collections_by_output_name(oper_temp_output)
        air_temp = sql_obj.data_collections_by_output_name(air_temp_output)
        rad_temp = sql_obj.data_collections_by_output_name(rad_temp_output)
        rel_humidity = sql_obj.data_collections_by_output_name(rel_humidity_output)

    else:  # use the honeybee_energy CLI
        # sqlite3 module doesn't work in Mac IronPython
        # or the file's big and we know that the Python3 version scales better
        # Execute the honybee CLI to obtain the results via CPython
        cmds = [folders.python_exe_path, '-m', 'honeybee_energy', 'result',
                'data-by-outputs', _sql] + all_output
        use_shell = True if os.name == 'nt' else False
        process = subprocess.Popen(cmds, stdout=subprocess.PIPE, shell=use_shell)
        stdout = process.communicate()
        data_coll_dicts = json.loads(stdout[0])
        # get all of the results
        oper_temp = serialize_data(data_coll_dicts[0])
        air_temp = serialize_data(data_coll_dicts[1])
        rad_temp = serialize_data(data_coll_dicts[2])
        rel_humidity = serialize_data(data_coll_dicts[3])



# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Create a window gas gap material that corresponds to a layer in a window construction.
This material can be plugged into the "HB Window Construction" component.
-

    Args:
        _name_: Text to set the name for the material and to be incorporated into
            a unique material identifier.
        _thickness_: Number for the thickness of the air gap layer in
            meters. (Default: 0.0125 m).
        _gas_types_: A list of text describing the types of gas in the gap.
            Text must be one of the following: 'Air', 'Argon', 'Krypton', 'Xenon'.
            Default: ['Air']
        _gas_ratios_: A list of text describing the volumetric fractions of gas
            types in the mixture.  This list must align with the gas_types
            input list. Default: Equal amout of gases for each type.

    Returns:
        mat: A window gas gap material that describes a layer in a window construction
            and can be assigned to a Honeybee Window construction.
"""

ghenv.Component.Name = 'HB Window Gap Material'
ghenv.Component.NickName = 'GapMat'
ghenv.Component.Message = '1.6.0'
ghenv.Component.Category = 'HB-Energy'
ghenv.Component.SubCategory = '1 :: Constructions'
ghenv.Component.AdditionalHelpFromDocStrings = '0'

try:  # import the core honeybee dependencies
    from honeybee.typing import clean_and_id_ep_string, clean_ep_string
except ImportError as e:
    raise ImportError('\nFailed to import honeybee:\n\t{}'.format(e))

try:  # import the honeybee-energy dependencies
    from honeybee_energy.material.gas import EnergyWindowMaterialGas, \
        EnergyWindowMaterialGasMixture
except ImportError as e:
    raise ImportError('\nFailed to import honeybee_energy:\n\t{}'.format(e))

try:  # import ladybug_rhino dependencies
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component):
    # set the default material properties
    _thickness_ = 0.0125 if _thickness_ is None else _thickness_
    _gas_types_ = ['Air'] if len(_gas_types_) == 0 else _gas_types_
    _gas_ratios_ = [1 / len(_gas_types_)] * len(_gas_types_) if \
        len(_gas_ratios_) == 0 else _gas_ratios_
    assert len(_gas_types_) == len(_gas_ratios_), \
        'Length of _gas_types_ does not equal length of _gas_ratios_.'
    name = clean_and_id_ep_string('GapMaterial') if _name_ is None else \
        clean_ep_string(_name_)

    # create the material
    if len(_gas_types_) == 1:
        mat = EnergyWindowMaterialGas(name, _thickness_, _gas_types_[0])
    else:
        mat = EnergyWindowMaterialGasMixture(name, _thickness_, _gas_types_, _gas_ratios_)
    if _name_ is not None:
        mat.display_name = _name_



# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Get information about carbon emission intensity (CEI) from an EnergyPlus SQL file.
_
The location and year (or input emissions of electricity intensity) will be used
to compute carbon intensity for both electricity and district heating/cooling.
Fixed numbers will be used to convert the following on-site fuel sources:
_
* Natural Gas --  277.358 kg/MWh
* Propane -- 323.897 kg/MWh
* Fuel Oil -- 294.962 kg/MWh
-

    Args:
        _sql: The file path of the SQL result file that has been generated from
            an energy simulation. This can also be a list of EnergyPlus files
            in which case EUI will be computed across all files. Lastly, it can
            be a directory or list of directories containing results, in which
            case, EUI will be calculated form all files ending in .sql.
        _loc_kgMWh: A ladybug Location object in the USA, which will be used to determine the
            subregion of the electrical grid. Alternatively, it can be A number
            for the electric grid carbon emissions in kg CO2/MWh. The following
            rules of thumb may be used as a guide:
            _
            * 800 kg/MWh - an inefficient coal or oil-dominated grid (West Virgina in 2020)
            * 400 kg/MWh - the average US (energy mixed) grid around 2020
            * 200-400 kg/MWh - for grids in transition to renewables
            * 100-200 kg/MWh - for grids with majority renewable/nuclear composition
            * 0-100 kg/MWh - for grids with renewables and storage
        _year_: An integer for the future year for which carbon emissions will
            be estimated. Values must be an even number and be between 2020
            and 2050. (Default: 2030).

    Returns:
        cei: A number for the total annual carbon emission intensity (CEI). This is
            the sum of all operational carbon emissions divided by the gross
            floor area (including both conditioned and unconditioned spaces).
            Units are kg CO2/m2.
        cei_end_use: The carbon emission intensity broken down by each end use. These
            values coorespond to the end_uses output below. Values are in kg CO2/m2 .
        end_uses: A list of text for each of the end uses in the simulation (Heating,
            Cooling, etc.). Thes outputs coorespond to the eui_end_use
            output above.
        gross_floor: The total gross floor area of the energy model. This can be used
            to compute the total energy use from the intensity values above or
            it can be used to help with other result post-processing. The value
            will be in m2 if ip_ is False or None and ft2 if True.
"""

ghenv.Component.Name = 'HB Carbon Emission Intensity'
ghenv.Component.NickName = 'CEI'
ghenv.Component.Message = '1.6.1'
ghenv.Component.Category = 'HB-Energy'
ghenv.Component.SubCategory = '6 :: Result'
ghenv.Component.AdditionalHelpFromDocStrings = '0'

import os
import subprocess
import json

try:
    from ladybug.location import Location
except ImportError as e:
    raise ImportError('\nFailed to import location:\n\t{}'.format(e))

try:
    from honeybee.config import folders
except ImportError as e:
    raise ImportError('\nFailed to import honeybee:\n\t{}'.format(e))

try:
    from honeybee_energy.result.emissions import future_electricity_emissions, \
        emissions_from_sql
except ImportError as e:
    raise ImportError('\nFailed to import honeybee_energy:\n\t{}'.format(e))

try:
    from ladybug_rhino.grasshopper import all_required_inputs, give_warning
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


# Use the SQLiteResult class to parse the result files directly on Windows.
def get_results_windows(sql_files, elec_emiss):
    results = emissions_from_sql(sql_files, elec_emiss)
    return results['carbon_intensity'], results['total_floor_area'], \
        results['end_uses'], results['sources']


# The SQLite3 module doesn't work in IronPython on Mac, so we must make a call
# to the Honeybee CLI (which runs on CPython) to get the results.
def get_results_mac(sql_files, elec_emiss):
    from collections import OrderedDict
    cmds = [folders.python_exe_path, '-m', 'honeybee_energy', 'result',
            'carbon-emission-intensity']
    cmds.extend(sql_files)
    cmds.extend(['--electricity-emissions', str(elec_emiss)])
    process = subprocess.Popen(cmds, stdout=subprocess.PIPE)
    stdout = process.communicate()
    results = json.loads(stdout[0], object_pairs_hook=OrderedDict)
    return results['carbon_intensity'], results['total_floor_area'], \
        results['end_uses'], results['sources']


if all_required_inputs(ghenv.Component):
    # ensure that _sql is a list rather than a single string
    if isinstance(_sql, basestring):
        _sql = [_sql]

    # process the location and year or the electricity intensity
    if isinstance(_loc_kgMWh, Location):
        yr = 2030 if _year_ is None else int(_year_)
        elec_emiss = future_electricity_emissions(_loc_kgMWh, yr)
        if elec_emiss is None:
            msg = 'Location must be inside the USA in order to be used for carbon ' \
                'emissions estimation.\nPlug in a number for carbon intensity in ' \
                'kg CO2/MWH for locations outside the USA.'
            print(msg)
            raise ValueError(msg)
    else:
        try:
            elec_emiss = float(_loc_kgMWh)
        except TypeError:
            msg = 'Expected location object or number for _loccation. ' \
                'Got {}.'.format(type(_loc_kgMWh))
            raise ValueError(msg)

    # get the results
    get_results = get_results_windows if os.name == 'nt' else get_results_mac
    cei, gross_floor, end_use_pairs, sources = get_results(_sql, elec_emiss)

    # create separate lists for end use values and labels
    cei_end_use = end_use_pairs.values()
    end_uses = [use.replace('_', ' ').title() for use in end_use_pairs.keys()]

    # give a warning if the sources include district heating or cooling
    if 'district_heat' in sources:
        msg = 'District heating was found in the results and so carbon emissions ' \
            'cannot be accurately estimated.\nTry using a different HVAC or SHW system.'
        print(msg)
        give_warning(ghenv.Component, msg)
    if 'district_cool' in sources:
        msg = 'District cooling was found in the results and so carbon emissions ' \
            'cannot be accurately estimated.\nTry using a different HVAC system.'
        print(msg)
        give_warning(ghenv.Component, msg)



# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Create a list of exterior subface (apertures + doors) constructions that can be
used to edit or create a ConstructionSet object.
-

    Args:
        _window_: A construction object for apertures with an Outdoors boundary
            condition and a Wall face type for their parent face. This can also
            be text for the identifier of the construction within the library.
        _skylight_: A construction object for apertures with an Outdoors boundary
            condition and a RoofCeiling or Floor face type for their parent face.
            This can also be text for the identifier of the construction within
            the library.
        _operable_: A construction object for apertures with an Outdoors boundary
            condition and True is_operable property. This can also be text for
            the identifier of the construction within the library.
        _exterior_door_: A construction object for opaque doors with an Outdoors
            boundary condition and a Wall face type for their parent face. This
            can also be text for the identifier of the construction within
            the library.
        _overhead_door_: A construction object for opaque doors with an Outdoors
            boundary condition and a RoofCeiling or Floor face type for their
            parent face. This can also be text for the identifier of the construction
            within the library.
        _glass_door_: A construction object for all glass doors with an Outdoors
            boundary condition. This can also be text for the identifier of the
            construction within the library.

    Returns:
        subface_set: A list of exterior subface constructions that can be used
            to edit or create a ConstructionSet object.
"""

ghenv.Component.Name = 'HB Subface Subset'
ghenv.Component.NickName = 'SubfaceSubset'
ghenv.Component.Message = '1.6.0'
ghenv.Component.Category = 'HB-Energy'
ghenv.Component.SubCategory = '0 :: Basic Properties'
ghenv.Component.AdditionalHelpFromDocStrings = '4'

try:  # import honeybee_energy dependencies
    from honeybee_energy.construction.opaque import OpaqueConstruction
    from honeybee_energy.construction.window import WindowConstruction
    from honeybee_energy.construction.windowshade import WindowConstructionShade
    from honeybee_energy.construction.dynamic import WindowConstructionDynamic
    from honeybee_energy.lib.constructions import opaque_construction_by_identifier, \
        window_construction_by_identifier
except ImportError as e:
    raise ImportError('\nFailed to import honeybee_energy:\n\t{}'.format(e))


def opaque_constr(construction, input_name):
    """Get an OpaqueConstrucion from the library if it's a string."""
    if isinstance(construction, str):
        return opaque_construction_by_identifier(construction)
    else:
        assert isinstance(construction, OpaqueConstruction), \
            'Expected OpaqueConstruction for {}. Got {}'.format(
                input_name, type(construction))
    return construction


def window_constr(construction, input_name):
    """Get an WindowConstrucion from the library if it's a string."""
    if isinstance(construction, str):
        return window_construction_by_identifier(construction)
    else:
        win_cls = (WindowConstruction, WindowConstructionShade, WindowConstructionDynamic)
        assert isinstance(construction, win_cls), 'Expected WindowConstruction ' \
            'for {}. Got {}'.format(input_name, type(construction))
    return construction


# go through each input construction
if _window_ is not None:
    _window_ = window_constr(_window_, '_window_')
if _skylight_ is not None:
    _skylight_ = window_constr(_skylight_, '_skylight_')
if _operable_ is not None:
    _operable_ = window_constr(_operable_, '_operable_')
if _exterior_door_ is not None:
    _exterior_door_ = opaque_constr(_exterior_door_, '_exterior_door_')
if _overhead_door_ is not None:
    _overhead_door_ = opaque_constr(_overhead_door_, '_overhead_door_')
if _glass_door_ is not None:
    _glass_door_ = window_constr(_glass_door_, '_glass_door_')


# return the final list from the component
subface_set = [_window_, _skylight_, _operable_, _exterior_door_,
               _overhead_door_, _glass_door_]



# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Apply a transmittance schedule to Honeybee Shade objects. Alternatively, it can
assign a transmittance schedule to all of the child shades of an Aperture, Door,
Face, or a Room.
_
This component supports the assigning of different schedules based on cardinal
orientation, provided that a list of transmittance schedule are input to
the _trans_sch. 
-

    Args:
        _hb_objs: Honeybee Shades, Apertures, Door, Faces, or Rooms to which the
            input _trans_sch should be assigned. For the case of a Honeybee Aperture,
            Door, Face or Room, the ShadeConstruction will be assigned to only the
            child shades directly assigned to that object. So passing in a Room
            will not change the schedule of shades assigned to Apertures
            of the Room's Faces. If this is the desired outcome, then the Room
            should be deconstructed into its child objects before using
            this component.
        _trans_sch: A Honeybee ScheduleRuleset or ScheduleFixedInterval to be
            applied to the input _hb_objs. This can also be text for a schedule
            to be looked up in the shade schedule library. If an array of text
            or schedule objects are input here, different schedules will be
            assigned based on cardinal direction, starting with north and
            moving clockwise.
    
    Returns:
        hb_objs: The input honeybee objects with their shade transmittance
            schedules edited.
"""

ghenv.Component.Name = "HB Apply Shade Schedule"
ghenv.Component.NickName = 'ApplyShadeSch'
ghenv.Component.Message = '1.6.0'
ghenv.Component.Category = 'HB-Energy'
ghenv.Component.SubCategory = '2 :: Schedules'
ghenv.Component.AdditionalHelpFromDocStrings = '3'


try:  # import the honeybee-energy extension
    from honeybee_energy.lib.schedules import schedule_by_identifier
except ImportError as e:
    raise ImportError('\nFailed to import honeybee_energy:\n\t{}'.format(e))

try:  # import the core honeybee dependencies
    from honeybee.shade import Shade
    from honeybee.room import Room
    from honeybee.face import Face
    from honeybee.aperture import Aperture
    from honeybee.door import Door
    from honeybee.orientation import angles_from_num_orient, face_orient_index
except ImportError as e:
    raise ImportError('\nFailed to import honeybee:\n\t{}'.format(e))

try:  # import the ladybug_rhino dependencies
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component):
    # duplicate the initial objects
    hb_objs = [obj.duplicate() for obj in _hb_objs]

    # process the input schedule
    for i, sch in enumerate(_trans_sch):
        if isinstance(sch, str):
            _trans_sch[i] = schedule_by_identifier(sch)

    # error message for unrecognized object
    error_msg = 'Input _hb_objs must be a Room, Face, Aperture, Door, or Shade. Not {}.'

    # assign the schedules
    if len(_trans_sch) == 1:
        for obj in hb_objs:
            if isinstance(obj, Shade):
                obj.properties.energy.transmittance_schedule = _trans_sch[0]
            elif isinstance(obj, (Aperture, Face, Room, Door)):
                for shd in obj.shades:
                    shd.properties.energy.transmittance_schedule = _trans_sch[0]
            else:
                raise TypeError(error_msg.format(type(obj)))
    else:  # assign schedules based on cardinal direction
        angles = angles_from_num_orient(len(_trans_sch))
        for obj in hb_objs:
            if isinstance(obj, (Aperture, Face, Door)):
                orient_i = face_orient_index(obj, angles)
                if orient_i is not None:
                    for shd in obj.shades:
                        shd.properties.energy.transmittance_schedule = _trans_sch[orient_i]
            elif isinstance(obj, Shade):
                obj.properties.energy.transmittance_schedule = _trans_sch[0]
            elif isinstance(obj, Room):
                 for shd in obj.shades:
                    shd.properties.energy.transmittance_schedule = _trans_sch[0]
            else:
                raise TypeError(error_msg.format(type(obj)))



# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Apply OpaqueConstruction to Honeybee Faces, Doors or Room walls.
_
This component supports the assigning of different constructions based on cardinal
orientation, provided that a list of OpaqueConstructions are input to the _constr. 
-

    Args:
        _hb_objs: Honeybee Faces, Doors or Rooms to which the input _constr should
            be assigned. For the case of a Honeybee Room, the construction
            will only be applied to the Room's outdoor walls. Note that, if you
            need to assign a construction to all the roofs, floors, etc. of a
            Room, the best practice is to create a ConstructionSet and assing that
            to the Room.
        _constr: A Honeybee OpaqueConstruction to be applied to the input _hb_objs.
            This can also be text for a construction to be looked up in the opaque
            construction library. If an array of text or construction objects
            are input here, different constructions will be assigned based on
            cardinal direction, starting with north and moving clockwise.
    
    Returns:
        hb_objs: The input honeybee objects with their constructions edited.
"""

ghenv.Component.Name = "HB Apply Opaque Construction"
ghenv.Component.NickName = 'ApplyOpaqueConstr'
ghenv.Component.Message = '1.6.0'
ghenv.Component.Category = 'HB-Energy'
ghenv.Component.SubCategory = '1 :: Constructions'
ghenv.Component.AdditionalHelpFromDocStrings = '3'


try:  # import the honeybee-energy extension
    from honeybee_energy.lib.constructions import opaque_construction_by_identifier
except ImportError as e:
    raise ImportError('\nFailed to import honeybee_energy:\n\t{}'.format(e))

try:  # import the core honeybee dependencies
    from honeybee.boundarycondition import Outdoors
    from honeybee.facetype import Wall
    from honeybee.room import Room
    from honeybee.face import Face
    from honeybee.door import Door
    from honeybee.orientation import angles_from_num_orient, face_orient_index
except ImportError as e:
    raise ImportError('\nFailed to import honeybee:\n\t{}'.format(e))

try:  # import the ladybug_rhino dependencies
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


def is_exterior_wall(face):
    """Check whether a given Face is an exterior Wall."""
    return isinstance(face.boundary_condition, Outdoors) and \
        isinstance(face.type, Wall)


if all_required_inputs(ghenv.Component):
    # duplicate the initial objects
    hb_objs = [obj.duplicate() for obj in _hb_objs]
    
    # process the input constructions
    for i, constr in enumerate(_constr):
        if isinstance(constr, str):
            _constr[i] = opaque_construction_by_identifier(constr)
    
    # error message for unrecognized object
    error_msg = 'Input _hb_objs must be a Room, Face, or Door. Not {}.'
    
    # assign the constructions
    if len(_constr) == 1:  # assign indiscriminately, even if it's horizontal
        for obj in hb_objs:
            if isinstance(obj, (Face, Door)):
                obj.properties.energy.construction = _constr[0]
            elif isinstance(obj, Room):
                for face in obj.faces:
                    if is_exterior_wall(face):
                        face.properties.energy.construction = _constr[0]
            else:
                raise TypeError(error_msg.format(type(obj)))
    else:  # assign constructions based on cardinal direction
        angles = angles_from_num_orient(len(_constr))
        for obj in hb_objs:
            if isinstance(obj, (Face, Door)):
                orient_i = face_orient_index(obj, angles)
                if orient_i is not None:
                    obj.properties.energy.construction = _constr[orient_i]
            elif isinstance(obj, Room):
                 for face in obj.faces:
                     if is_exterior_wall(face):
                         orient_i = face_orient_index(face, angles)
                         if orient_i is not None:
                            face.properties.energy.construction = _constr[orient_i]
            else:
                raise TypeError(error_msg.format(type(obj)))

